﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Transformations;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class SymbolicExpressionTreeBacktransformator : IModelBacktransformator {
    private readonly ITransformationMapper<ISymbolicExpressionTreeNode> transformationMapper;

    public SymbolicExpressionTreeBacktransformator(ITransformationMapper<ISymbolicExpressionTreeNode> transformationMapper) {
      this.transformationMapper = transformationMapper;
    }

    public void Backtransform(IDataAnalysisModel model, IEnumerable<ITransformation> transformations) {
      var symbolicModel = (ISymbolicDataAnalysisModel)model;

      foreach (var transformation in transformations) {
        ApplyBacktransformation(transformation, symbolicModel.SymbolicExpressionTree);
      }
    }

    private void ApplyBacktransformation(ITransformation transformation, ISymbolicExpressionTree symbolicExpressionTree) {
      var variableNodes = symbolicExpressionTree.IterateNodesBreadth()
        .OfType<VariableTreeNode>()
        .Where(n => n.VariableName == transformation.Column);

      var transformationTree = transformationMapper.GenerateModel(transformation);
      foreach (var variableNode in variableNodes) {
        SwapTransformationWithVariable(transformationTree, variableNode);
      }
    }

    private void SwapTransformationWithVariable(ISymbolicExpressionTreeNode transformationTree, VariableTreeNode variableNode) {
      var parent = variableNode.Parent;
      int index = parent.IndexOfSubtree(variableNode);
      parent.RemoveSubtree(index);

      var multiplicationNode = new SymbolicExpressionTreeNode(new Multiplication());
      multiplicationNode.AddSubtree(new ConstantTreeNode(new Constant()) { Value = variableNode.Weight });
      multiplicationNode.AddSubtree(transformationTree);

      parent.InsertSubtree(index, multiplicationNode);
    }
  }
}
